@@ -16,21 +16,23 @@ from functools import lru_cache
 from typing import (
     Any,
     List,
+    Union,
 )
 
 import inflection
 from gql import Client
-from graphql import (
-    GraphQLNonNull,
-    GraphQLType,
-)
+from graphql import GraphQLType
 from tqdm.auto import tqdm
 
 from ..queries import (
+    MultiRefreshActionQuery,
     RefreshActionQuery,
     StartActionQuery,
 )
-from .async_result import AsyncResult
+from .async_result import (
+    AsyncResult,
+    AsyncResultCollection,
+)
 from .custom_handlers import (
     ComplexArrayEntryHandler,
     ComplexArrayHandler,
@@ -98,64 +100,42 @@ class QctrlGraphQLEnvironment(GraphQLEnvironment):
         data: Any,
     ) -> StartActionQuery:
         """Returns the corresponding StartActionQuery object."""
-        payload = self.format_payload(input_type, data)
-        selection = self.get_selection(result_type, self._action_selection_override)
-
         query = StartActionQuery(
-            self._gql_client,
-            mutation_name=mutation_name,
-            payload=payload,
-            selection=selection,
+            self._gql_client, self, mutation_name, input_type, result_type, data
         )
 
         return query
 
     def wait_for_completion(
-        self, field_type: GraphQLType, obj: Any, refresh_query: RefreshActionQuery
+        self,
+        refresh_query: Union[RefreshActionQuery, MultiRefreshActionQuery],
+        *async_results: AsyncResult
     ):
-        """Waits until the corresponding action is completed on the server."""
+        """Waits until the corresponding actions are completed on the server."""
         _wait = Wait()
 
         with tqdm(total=100, leave=False) as progress_bar:
-            _total = 0
-            async_result = AsyncResult(obj, progress_bar, refresh_query)
+            promise = AsyncResultCollection(*async_results)
+
+            while not promise.is_completed:
+
+                promise.update_progress(progress_bar)
 
-            while not obj.is_completed:
-                progress = int(obj.progress * 100) - _total
-                _total += progress
-                progress_bar.update(progress)
                 _wait()
-                result = async_result.refresh()
-                self.update_data(field_type, obj, result)
-                async_result.write_waiting_message()
+
+                # check completion
+                response = refresh_query(*promise.get_action_ids())
+                promise.update_data(self, response, progress_bar)
 
             progress_bar.update(max(0, progress_bar.total - progress_bar.n))
 
-        # action is now completed
-        if not obj.is_successful:
-            if obj.job_errors:
-                raise RuntimeError(f"Execution failed: {obj.job_errors.exception}")
-
-            if obj.status == obj.FAILURE:
-                raise RuntimeError(
-                    "Execution resulted with status 'FAILURE'("
-                    "no extra details about the errors are "
-                    "available)."
-                )
-
-        LOGGER.info(
-            "Action %s finished with status: %s", str(obj.action_id), obj.status
-        )
+        promise.finalize()
 
     def build_refresh_query(self, field_type: GraphQLType) -> RefreshActionQuery:
         """Returns a RefreshActionQuery object for
         refreshing an object populated with response data.
         """
-
-        if isinstance(field_type, GraphQLNonNull):
-            field_type = field_type.of_type
-
-        return RefreshActionQuery(self._gql_client, field_type=field_type.name)
+        return RefreshActionQuery(self._gql_client, field_type)
 
 
 def create_environment(gql_client: Client) -> QctrlGraphQLEnvironment:
